#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pathlib import Path
from subprocess import CalledProcessError
from tempfile import TemporaryDirectory, NamedTemporaryFile
from unittest import TestCase
from unittest.mock import NonCallableMock, call, patch, DEFAULT
from winreg import HKEY_LOCAL_MACHINE, HKEY_CURRENT_USER, KEY_WOW64_32KEY, KEY_READ, KEY_WOW64_64KEY

import thespiae.install.system
from thespiae.install.data import DownloadSpec, ContentRangeData
from thespiae.install.exception import HashError, UnsupportedResponseError
from thespiae.install.protocol import Feedback
from thespiae.install.system import parse_content_range
from .helper import TestWebServer, copy_to, ExtraAssertMixin, assert_download_progress_for, create_archive
from .singleton import real_download_manager as download_manager, real_install_manager as install_manager


def setUpModule():
    global web_server
    web_server = TestWebServer()
    web_server.start()


def tearDownModule():
    web_server.stop()


class TestContentRangeParser(TestCase):

    def test_content_range_parsing(self):
        self.assertEqual(parse_content_range('bytes */123'), ContentRangeData(None, None, 123))

    def test_content_range_parsing2(self):
        self.assertEqual(parse_content_range('bytes */*'), ContentRangeData(None, None, None))

    def test_content_range_parsing3(self):
        self.assertEqual(parse_content_range('bytes 12-23/34'), ContentRangeData(12, 23, 34))

    def test_content_range_parsing4(self):
        self.assertIsNone(parse_content_range('bytes 12-23/'))

    def test_content_range_parsing5(self):
        self.assertIsNone(parse_content_range('/'))

    def test_content_range_parsing6(self):
        self.assertIsNone(parse_content_range(''))

    def test_content_range_parsing7(self):
        self.assertEqual(parse_content_range('bytes 12-23/*'), ContentRangeData(12, 23, None))

    def test_content_range_parsing8(self):
        self.assertIsNone(parse_content_range('bytes 12-/23'))

    def test_content_range_parsing9(self):
        self.assertIsNone(parse_content_range('bytes -12/23'))


class TestInstallManager(ExtraAssertMixin, TestCase):

    def test_file_copy_and_removal(self):
        with TemporaryDirectory() as temp_path:
            with NamedTemporaryFile(dir=temp_path, delete=False) as temp_file:
                temp_file.close()
                tp = Path(temp_path) / 'test'
                path = tp / 'test2'
                temp_file_path = Path(temp_file.name)
                copied_file_path = path / temp_file_path.name
                install_manager.copy_file(str(temp_file.name), path, temp_file_path.name)
                self.assertTrue(copied_file_path.exists())
                install_manager.remove_file(path, temp_file_path.name)
                self.assertFalse(copied_file_path.exists())
                self.assertFalse(tp.exists())
                self.assertFalse(path.exists())

    def test_archive_unpacking_and_removal(self):
        with TemporaryDirectory() as temp_path:
            base_path = Path(temp_path) / 'test_archive'
            unpack_path = base_path.parent / 'unpack_dir' / 'unpack_dir2'
            archive_path = create_archive(base_path, 'bztar')
            install_manager.unpack_archive(archive_path, 'bztar', unpack_path)
            self.assertTrue(unpack_path.exists())
            self.assertTrue(unpack_path / 'test.txt')
            self.assertTrue(unpack_path / 'test2.txt')
            self.assertTrue(unpack_path / 'test4.txt')
            install_manager.remove_directory(unpack_path)
            self.assertFalse(unpack_path.parent.exists())

    def test_file_presence(self):
        with NamedTemporaryFile() as file:
            self.assertTrue(install_manager.is_file_present(file.name))
            file.close()
            self.assertFalse(install_manager.is_file_present(file.name))

    @patch.multiple(thespiae.install.system, subprocess_run=DEFAULT)
    def test_running_file(self, subprocess_run):
        install_manager.run_file('test_path.exe', ['arg1', 'arg2'])
        subprocess_run.assert_called_once_with(['test_path.exe', 'arg1', 'arg2'], check=True, shell=True)

    @patch.multiple(thespiae.install.system, subprocess_run=DEFAULT)
    def test_running_file2(self, subprocess_run):
        install_manager.run_file('test_path.exe', None)
        subprocess_run.assert_called_once_with(['test_path.exe'], check=True, shell=True)

    @patch.multiple(thespiae.install.system, subprocess_run=DEFAULT)
    def test_installing_msi(self, subprocess_run):
        install_manager.install_msi('test_package.msi', ['arg1', 'arg2'])
        subprocess_run.assert_called_once_with(['msiexec', '/I', 'test_package.msi', 'arg1', 'arg2'],
                                               capture_output=True, check=True, shell=True, text=True)

    @patch.multiple(thespiae.install.system, subprocess_run=DEFAULT)
    def test_installing_msi2(self, subprocess_run):
        install_manager.install_msi('test_package.msi', None)
        subprocess_run.assert_called_once_with(['msiexec', '/I', 'test_package.msi'],
                                               capture_output=True, check=True, shell=True, text=True)

    @patch.multiple(thespiae.install.system, subprocess_run=DEFAULT)
    def test_running_command(self, subprocess_run):
        install_manager.run_command('test_command', ['arg1', 'arg2'])
        subprocess_run.assert_called_once_with(['test_command', 'arg1', 'arg2'],
                                               capture_output=True, check=True, shell=True, text=True)

    @patch.multiple(thespiae.install.system, subprocess_run=DEFAULT)
    def test_running_command2(self, subprocess_run):
        install_manager.run_command('test_command', None)
        subprocess_run.assert_called_once_with(['test_command'],
                                               capture_output=True, check=True, shell=True, text=True)

    @patch.multiple(thespiae.install.system, subprocess_run=DEFAULT)
    def test_running_command_issue(self, subprocess_run):
        subprocess_run.side_effect = CalledProcessError(-1, ['test_command', 'arg1', 'arg2'])
        with self.assertRaises(CalledProcessError):
            install_manager.run_command('test_command', ['arg1', 'arg2'])
        subprocess_run.assert_called_once_with(['test_command', 'arg1', 'arg2'],
                                               capture_output=True, check=True, shell=True, text=True)

    @patch.multiple(thespiae.install.system, subprocess_run=DEFAULT)
    def test_running_command_no_restart(self, subprocess_run):
        subprocess_run.side_effect = CalledProcessError(3010, ['test_command', 'arg1', 'arg2'])
        install_manager.run_command('test_command', ['arg1', 'arg2'])
        subprocess_run.assert_called_once_with(['test_command', 'arg1', 'arg2'],
                                               capture_output=True, check=True, shell=True, text=True)

    @patch.multiple(thespiae.install.system, OpenKeyEx=DEFAULT)
    def test_product_code_presence(self, OpenKeyEx):
        err = OSError()
        err.winerror = 2
        OpenKeyEx.side_effect = err
        self.assertFalse(install_manager.is_product_code_present('{123}'))
        self.assert_called_exactly_with_no_order(OpenKeyEx,
                                                 call(HKEY_LOCAL_MACHINE,
                                                      r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\{123}', 0,
                                                      KEY_READ | KEY_WOW64_32KEY),
                                                 call(HKEY_LOCAL_MACHINE,
                                                      r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\{123}', 0,
                                                      KEY_READ | KEY_WOW64_64KEY),
                                                 call(HKEY_CURRENT_USER,
                                                      r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\{123}', 0,
                                                      KEY_READ | KEY_WOW64_32KEY),
                                                 call(HKEY_CURRENT_USER,
                                                      r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\{123}', 0,
                                                      KEY_READ | KEY_WOW64_64KEY))

    @patch.multiple(thespiae.install.system, OpenKeyEx=DEFAULT)
    def test_product_code_presence2(self, OpenKeyEx):
        self.assertTrue(install_manager.is_product_code_present('{123}'))


class TestDownloadManager(ExtraAssertMixin, TestCase):

    def test_spec_filtering(self):
        with TemporaryDirectory() as temp_path:
            p = Path(temp_path) / 'test.txt'
            p2 = Path(temp_path) / 'test2.txt'
            sp = DownloadSpec(url=web_server.create_url_for('static/test.txt'),
                              hash='bddace389020162bfc3cec0d82042a8eeebc73cb0fdcf9b50bde03a476329d66',
                              download_path=str(p), part_path=str(p) + '.part', name='a', version='1')
            sp2 = DownloadSpec(url=web_server.create_url_for('static/test2.txt'), hash=None, download_path=str(p2),
                               part_path=str(p2) + '.part', name='a', version='2')
            copy_to('test.txt', str(p))
            self.assertSequenceEqual(download_manager.filter_present([sp, sp2]), [sp2])

    def test_spec_filtering2(self):
        with TemporaryDirectory() as temp_path:
            p = Path(temp_path) / 'test.txt'
            p2 = Path(temp_path) / 'test2.txt'
            sp = DownloadSpec(url=web_server.create_url_for('static/test.txt'),
                              hash='bddace389020162bfc3cec0d82042a8eeebc73cb0fdcf9b50bde03a476329d66',
                              download_path=str(p), part_path=str(p) + '.part', name='a', version='1')
            sp2 = DownloadSpec(url=web_server.create_url_for('static/test2.txt'), hash=None, download_path=str(p2),
                               part_path=str(p2) + '.part', name='a', version='2')
            copy_to('test.txt', str(p), 5)
            with self.assertRaises(HashError) as c:
                download_manager.filter_present([sp, sp2])
            self.assertEqual(c.exception.download_spec, sp)
            self.assertEqual(c.exception.actual_hash,
                             '5994471abb01112afcc18159f6cc74b4f511b99806da59b3caf5a9c173cacfc5')

    def test_no_check_downloading_feedback(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test.txt'
            p2 = Path(temp_path) / 'test2.txt'
            sp = DownloadSpec(url=web_server.create_url_for('static/test.txt'), hash=None, download_path=str(p),
                              part_path=str(p) + '.part', name='a', version='1')
            sp2 = DownloadSpec(url=web_server.create_url_for('static/test2.txt'), hash=None, download_path=str(p2),
                               part_path=str(p2) + '.part', name='a', version='2')
            download_manager.download([sp, sp2], fb)
            self.assertTrue(p.exists())
            self.assertTrue(p2.exists())
            fb.report_download_started.assert_called_once()
            self.assert_called_exactly_with_no_order(fb.report_entry_download_initiated, call(sp), call(sp2))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_started, call(sp, 0, 18),
                                                     call(sp2, 0, 18))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_finished, call(sp), call(sp2))
            assert_download_progress_for(fb, {sp, sp2})
            fb.report_download_finished.assert_called_once()

    def test_hash_check_downloading(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test.txt'
            p2 = Path(temp_path) / 'test2.txt'
            sp = DownloadSpec(url=web_server.create_url_for('static/test.txt'),
                              hash='bddace389020162bfc3cec0d82042a8eeebc73cb0fdcf9b50bde03a476329d66',
                              download_path=str(p), part_path=str(p) + '.part', name='a', version='1')
            sp2 = DownloadSpec(url=web_server.create_url_for('static/test2.txt'),
                               hash='a132b1e5a4bd5a4f201e2b3fffdefaf7a9864dcae43c1963eeab7c13661ae6be',
                               download_path=str(p2), part_path=str(p2) + '.part', name='b', version='2')
            download_manager.download([sp, sp2], fb)
            self.assertTrue(p.exists())
            self.assertTrue(p2.exists())
            fb.report_download_started.assert_called_once()
            self.assert_called_exactly_with_no_order(fb.report_entry_download_initiated, call(sp), call(sp2))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_started, call(sp, 0, 18),
                                                     call(sp2, 0, 18))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_finished, call(sp), call(sp2))
            assert_download_progress_for(fb, {sp, sp2})
            fb.report_download_finished.assert_called_once()

    def test_hash_check_downloading2(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test.txt'
            p2 = Path(temp_path) / 'test2.txt'
            sp = DownloadSpec(url=web_server.create_url_for('static/test.txt'),
                              hash='bddace389020162bfc3cec0d82042a8eeebc73cb0fdcf9b50bde03a476329d66',
                              download_path=str(p), part_path=str(p) + '.part', name='a', version='1')
            sp2 = DownloadSpec(url=web_server.create_url_for('static/test2.txt'), hash='123', download_path=str(p2),
                               part_path=str(p2) + '.path', name='b', version='2')
            with self.assertRaises(HashError) as c:
                download_manager.download([sp, sp2], fb)
            self.assertEqual(c.exception.download_spec, sp2)
            self.assertEqual(c.exception.actual_hash,
                             'a132b1e5a4bd5a4f201e2b3fffdefaf7a9864dcae43c1963eeab7c13661ae6be')
            fb.report_download_started.assert_called_once()
            self.assert_called_exactly_with_no_order(fb.report_entry_download_initiated, call(sp), call(sp2))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_started, call(sp, 0, 18),
                                                     call(sp2, 0, 18))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_finished, call(sp), call(sp2))
            assert_download_progress_for(fb, {sp, sp2})
            fb.report_download_finished.assert_not_called()

    def test_not_found_downloading(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test3.txt'
            p2 = Path(temp_path) / 'test' / 'test2.txt'
            sp = DownloadSpec(url=web_server.create_url_for('static/test3.txt'), hash=None, download_path=str(p),
                              part_path=str(p) + '.part', name='a', version='1')
            sp2 = DownloadSpec(url=web_server.create_url_for('static/test2.txt'),
                               hash='a132b1e5a4bd5a4f201e2b3fffdefaf7a9864dcae43c1963eeab7c13661ae6be',
                               download_path=str(p2), part_path=str(p2) + '.part', name='b', version='2')
            with self.assertRaises(UnsupportedResponseError) as c:
                download_manager.download([sp, sp2], fb)
            self.assertEqual(c.exception.download_spec, sp)
            self.assertEqual(c.exception.status, 404)
            self.assertIsNotNone(c.exception.headers)
            fb.report_download_started.assert_called_once()
            self.assert_called_exactly_with_no_order(fb.report_entry_download_initiated, call(sp), call(sp2))
            fb.report_download_finished.assert_not_called()

    def test_partial_downloading(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test.txt'
            p2 = Path(temp_path) / 'test2.txt'
            sp = DownloadSpec(url=web_server.create_url_for('static/test.txt'),
                              hash='bddace389020162bfc3cec0d82042a8eeebc73cb0fdcf9b50bde03a476329d66',
                              download_path=str(p), part_path=str(p) + '.part', name='a', version='1')
            sp2 = DownloadSpec(url=web_server.create_url_for('static/test2.txt'),
                               hash='a132b1e5a4bd5a4f201e2b3fffdefaf7a9864dcae43c1963eeab7c13661ae6be',
                               download_path=str(p2), part_path=str(p2) + '.part', name='b', version='2')
            copy_to('test2.txt', str(p2) + '.part', 12)
            download_manager.download([sp, sp2], fb)
            self.assertTrue(p.exists())
            self.assertTrue(p2.exists())
            self.assertCountEqual(web_server.get_request_data(),
                                  [('/static/test.txt', None), ('/static/test2.txt', 'bytes=12-')])
            fb.report_download_started.assert_called_once()
            self.assert_called_exactly_with_no_order(fb.report_entry_download_initiated, call(sp), call(sp2))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_started, call(sp, 0, 18),
                                                     call(sp2, 12, 18))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_finished, call(sp), call(sp2))
            assert_download_progress_for(fb, {sp, sp2})
            fb.report_download_finished.assert_called_once()

    def test_partial_downloading2(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test.txt'
            p2 = Path(temp_path) / 'test2.txt'
            sp = DownloadSpec(url=web_server.create_url_for('static/test.txt'),
                              hash='bddace389020162bfc3cec0d82042a8eeebc73cb0fdcf9b50bde03a476329d66',
                              download_path=str(p), part_path=str(p) + '.part', name='a', version='1')
            sp2 = DownloadSpec(url=web_server.create_url_for('static/test2.txt'),
                               file_hash='a132b1e5a4bd5a4f201e2b3fffdefaf7a9864dcae43c1963eeab7c13661ae6be',
                               download_path=str(p2), part_path=str(p2) + '.part', name='b', version='2')
            copy_to('test2.txt', str(p2) + '.part')
            download_manager.download([sp, sp2], fb)
            self.assertTrue(p.exists())
            self.assertTrue(p2.exists())
            self.assertCountEqual(web_server.get_request_data(),
                                  [('/static/test.txt', None), ('/static/test2.txt', 'bytes=18-')])
            fb.report_download_started.assert_called_once()
            self.assert_called_exactly_with_no_order(fb.report_entry_download_initiated, call(sp), call(sp2))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_started, call(sp, 0, 18),
                                                     call(sp2, 18, 18))
            self.assert_called_exactly_with_no_order(fb.report_entry_download_finished, call(sp), call(sp2))
            assert_download_progress_for(fb, {sp})
            fb.report_download_finished.assert_called_once()

    def test_partial_downloading3(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test4.txt'
            copy_to('test2.txt', str(p) + '.part', 10)
            sp = DownloadSpec(url=web_server.create_url_for('dynamic/test4.txt'),
                              hash='771c8bf33bbaea2ebc52d77cbd010185bc7ec53a0a44860cf8d3e352cebc4462',
                              download_path=str(p), part_path=str(p) + '.part', name='a', version='1')
            download_manager.download([sp], fb)
            self.assertTrue(p.exists())
            self.assertCountEqual(web_server.get_request_data(), [('/dynamic/test4.txt', 'bytes=10-')])
            fb.report_download_started.assert_called_once()
            fb.report_entry_download_initiated.assert_called_once_with(sp)
            fb.report_entry_download_started.assert_called_once_with(sp, 0, 56)
            fb.report_entry_download_finished.assert_called_once_with(sp)
            assert_download_progress_for(fb, {sp})
            fb.report_download_finished.assert_called_once()

    def test_partial_downloading4(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test5.txt'
            sp = DownloadSpec(url=web_server.create_url_for('dynamic/test5.txt'), hash=None, download_path=str(p),
                              part_path=str(p) + '.part', name='a', version='1')
            copy_to('test2.txt', str(p) + '.part', 10)
            with self.assertRaises(UnsupportedResponseError) as c:
                download_manager.download([sp], fb)
            self.assertEqual(c.exception.download_spec, sp, fb)
            self.assertEqual(c.exception.status, 416)
            self.assertIsNotNone(c.exception.headers)
            self.assertCountEqual(web_server.get_request_data(), [('/dynamic/test5.txt', 'bytes=10-')])
            fb.report_download_started.assert_called_once()
            fb.report_entry_download_initiated.assert_called_once_with(sp)
            fb.report_download_finished.assert_not_called()

    def test_partial_downloading5(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test6.txt'
            sp = DownloadSpec(url=web_server.create_url_for('dynamic/test6.txt'), hash=None, download_path=str(p),
                              part_path=str(p) + '.part', name='a', version='1')
            copy_to('test2.txt', str(p) + '.part', 10)
            with self.assertRaises(UnsupportedResponseError) as c:
                download_manager.download([sp], fb)
            self.assertEqual(c.exception.download_spec, sp, fb)
            self.assertEqual(c.exception.status, 416)
            self.assertIsNotNone(c.exception.headers)
            self.assertCountEqual(web_server.get_request_data(), [('/dynamic/test6.txt', 'bytes=10-')])
            fb.report_download_started.assert_called_once()
            fb.report_entry_download_initiated.assert_called_once_with(sp)
            fb.report_download_finished.assert_not_called()

    def test_skiping_downloaded_files(self):
        web_server.reset_request_data()
        with TemporaryDirectory() as temp_path:
            fb = NonCallableMock(spec_set=Feedback)
            p = Path(temp_path) / 'test.txt'
            p2 = Path(temp_path) / 'test2.txt'
            sp = DownloadSpec(url=web_server.create_url_for('static/test.txt'),
                              hash='bddace389020162bfc3cec0d82042a8eeebc73cb0fdcf9b50bde03a476329d66',
                              download_path=str(p), part_path=str(p) + '.part', name='a', version='1')
            sp2 = DownloadSpec(url=web_server.create_url_for('static/test2.txt'),
                               hash='a132b1e5a4bd5a4f201e2b3fffdefaf7a9864dcae43c1963eeab7c13661ae6be',
                               download_path=str(p2), part_path=str(p2) + '.part', name='b', version='2')
            copy_to('test2.txt', str(p2))
            download_manager.download(download_manager.filter_present([sp, sp2]), fb)
            self.assertTrue(p.exists())
            self.assertTrue(p2.exists())
            self.assertCountEqual(web_server.get_request_data(), [('/static/test.txt', None)])
            fb.report_download_started.assert_called_once()
            fb.report_entry_download_initiated.assert_called_once_with(sp)
            fb.report_entry_download_started.assert_called_once_with(sp, 0, 18)
            fb.report_entry_download_finished.assert_called_once_with(sp)
            assert_download_progress_for(fb, {sp})
            fb.report_download_finished.assert_called_once()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import setup, find_packages

setup(name='thespiae', version='0.1',
      python_requires='>=3.8.1',
      install_requires=['antlr4-python3-runtime~=4.7.2', 'pyyaml>=5.1', 'aiohttp>=3.6.2 ', 'tqdm>=4.31.1',
                        'colorama~=0.4.1'],
      packages=find_packages(exclude=['thespiae.tests', 'thespiae.tests.*']),
      entry_points={'console_scripts': ['thespiae=thespiae.core:main']},
      description='Tool to setup Windows-based development environments',
      url='https://github.com/apleshakov/thespiae',
      author='Andrey Pleshakov',
      author_email='aplshkv@gmail.com',
      long_description_content_type='text/x-rst',
      license='Apache-2.0',
      classifiers=['Environment :: Console',
                   'License :: OSI Approved :: Apache Software License',
                   'Natural Language :: English',
                   'Operating System :: Microsoft :: Windows',
                   'Programming Language :: Python :: 3.8',
                   'Topic :: Utilities'])



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

import sys
from argparse import ArgumentParser
from dataclasses import dataclass
from gettext import translation
from os import linesep
from typing import TYPE_CHECKING

from tqdm import tqdm

from .exception import ThespiaeError
from .install.protocol import Feedback as InstallFeedback
from .path.protocol import Feedback as PathFeedback

if TYPE_CHECKING:
    from typing import Sequence, MutableMapping
    from .conf import AppEntry, AppEntryRef
    from .install import DownloadSpec

t = translation('thespiae', fallback=True)
_ = t.gettext


def _info(o) -> None:
    print(o, file=sys.stdout)


def _error(o) -> None:
    print(o, file=sys.stderr)


def _confirm(prompt: str) -> None:
    # TODO: implement through tty <AP>
    resp = input(prompt + ' [y/N]:')
    if 'y' != resp:
        raise _NotConfirmedError


def _position_after_bars(bar_count: int) -> None:
    if bar_count > 0:
        sys.stdout.write('\r')
        sys.stdout.write('\x1b[{}B'.format(bar_count - 1))
        sys.stdout.flush()


class _NotConfirmedError(Exception):
    pass


class CLIFeedback(InstallFeedback, PathFeedback):

    def report_checking_software(self):
        _info(_('Checking software'))

    def __init__(self):
        self._download_name_max_len = 0
        self._download_bars: MutableMapping[AppEntryRef, tqdm] = {}
        self._uninstall_count_total = 0
        self._install_count_total = 0
        self._uninstall_count = 0
        self._install_count = 0

    def confirm_operations(self, to_download: Sequence[DownloadSpec], to_uninstall: Sequence[AppEntry],
                           to_install: Sequence[AppEntry]):
        # TODO: multi-line indent <AP>
        if to_download:
            self._download_bars = {}
            self._download_name_max_len = max((len(str(spec.ref)) for spec in to_download))
            _info(_('Download distribution for'))
            _info('\t' + ' '.join(str(e.ref) for e in to_download))
        if to_uninstall:
            self._uninstall_count_total = len(to_uninstall)
            self._uninstall_count = 0
            _info(_('To uninstall'))
            _info('\t' + ' '.join(str(e.ref) for e in to_uninstall))
        if to_install:
            self._install_count_total = len(to_install)
            self._install_count = 0
            _info(_('To install'))
            _info('\t' + ' '.join(str(e.ref) for e in to_install))
        _confirm(_('Continue?'))

    def report_software_set_no_changes(self):
        _info(_('Installed software matches requirements, no changes needed'))

    def report_download_started(self):
        _info(_('Downloading'))

    def report_entry_download_initiated(self, spec: DownloadSpec):
        self._download_bars[spec.ref] = tqdm(desc=str(spec.ref).rjust(self._download_name_max_len), unit='B',
                                             unit_scale=True, miniters=1000 * 1000, ascii=True)

    def report_entry_download_started(self, spec: DownloadSpec, current_size: int, total_size: int = None):
        bar = self._download_bars[spec.ref]
        if total_size:
            bar.total = total_size
        bar.n = current_size
        bar.refresh()

    def report_entry_download_progress(self, spec: DownloadSpec, batch_size: int):
        bar = self._download_bars[spec.ref]
        bar.update(batch_size)

    def report_entry_download_finished(self, spec: DownloadSpec):
        pass

    def report_download_finished(self):
        self._finalize_download()

    def report_removal_started(self):
        _info(_('Uninstalling software'))

    def report_entry_removal_started(self, entry: AppEntry):
        self._uninstall_count += 1
        _info(_('[{0}/{1}] uninstalling {2}').format(self._uninstall_count, self._uninstall_count_total, entry.ref))

    def report_entry_removal_finished(self, entry: AppEntry):
        _info(_('{} uninstalled').format(entry.ref))

    def report_removal_finished(self):
        pass

    def report_installation_started(self):
        _info(_('Installing software'))

    def report_entry_installation_started(self, entry: AppEntry):
        self._install_count += 1
        _info(_('[{0}/{1}] installing {2}').format(self._install_count, self._install_count_total, entry.ref))

    def report_entry_installation_finished(self, entry: AppEntry):
        _info(_('{} installed').format(entry.ref))

    def report_installation_finished(self):
        pass

    def report_path_analysis(self):
        _info(_('Checking user PATH'))

    def confirm_user_path_update(self, before: str, after):
        # TODO: multi-line indent <AP>
        _info(_('Current user PATH'))
        _info('\t' + before)
        _info(_('Updated user PATH'))
        _info('\t' + after)
        _confirm(_('Continue?'))

    def report_user_path_no_changes(self):
        _info(_('No changes to user PATH are needed'))

    def report_user_path_updated(self):
        _info(_('User PATH updated'))

    def finalize(self):
        self._finalize_download()

    def _finalize_download(self):
        bars = list(self._download_bars.values())
        for bar in bars:
            bar.refresh()
        for bar in reversed(bars):
            bar.close()
        _position_after_bars(len(bars))


@dataclass(frozen=True)
class CLIParams:
    config_file: str
    data_dir: str


class CLI:

    def __init__(self, app_name: str):
        parser = ArgumentParser(app_name)
        parser.add_argument('config_file', metavar=_('CONFIG_FILE'), help=_('YAML file to read configuration from'))
        parser.add_argument('data_dir', metavar=_('DATA_DIR'),
                            help=_('directory to download to and install software from'))
        ns = parser.parse_args()
        self._params = CLIParams(ns.config_file, ns.data_dir)
        self._feedback = CLIFeedback()

    @property
    def params(self) -> CLIParams:
        return self._params

    @property
    def feedback(self) -> CLIFeedback:
        return self._feedback

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            self.feedback.finalize()
            if issubclass(exc_type, ThespiaeError):
                _error(str(exc_val))
                if exc_val.__cause__:
                    _error(_('because of'))
                    _error(str(exc_val.__cause__))
                return True
            elif issubclass(exc_type, _NotConfirmedError) or issubclass(exc_type, KeyboardInterrupt):
                _error(linesep + _('Action cancelled by the user'))
                return True



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from asyncio import new_event_loop, set_event_loop, run_coroutine_threadsafe, all_tasks, gather
from functools import partial
from io import SEEK_END
from pathlib import Path
from shutil import make_archive
from threading import Thread, Event
from typing import TYPE_CHECKING
from urllib.parse import urlunsplit

from aiohttp.test_utils import unused_port
from aiohttp.web import Application, AppRunner, TCPSite, middleware, StreamResponse, Response, \
    HTTPRequestRangeNotSatisfiable

from .singleton import install_manager, download_manager

if TYPE_CHECKING:
    from typing import List, Set
    from unittest import TestCase

    from thespiae.install.data import DownloadSpec
    from thespiae.install.protocol import Feedback


class InstallManagerMockResetMixin:

    def setUp(self, *args, **kwargs):
        super().setUp(*args, **kwargs)
        install_manager.reset_mock(return_value=True, side_effect=True)
        install_manager.is_file_present.reset_mock(return_value=True, side_effect=True)
        install_manager.is_product_code_present.reset_mock(return_value=True, side_effect=True)
        install_manager.run_command.reset_mock(return_value=True, side_effect=True)
        install_manager.copy_file.reset_mock(return_value=True, side_effect=True)
        install_manager.remove_file.reset_mock(return_value=True, side_effect=True)
        install_manager.unpack_archive.reset_mock(return_value=True, side_effect=True)
        install_manager.remove_directory.reset_mock(return_value=True, side_effect=True)


class DownloadManagerMockResetMixin:

    def setUp(self, *args, **kwargs):
        super().setUp(*args, **kwargs)
        download_manager.reset_mock(return_value=True, side_effect=True)
        download_manager.filter_present.reset_mock(return_value=True, side_effect=True)


def set_install_manager_data(is_file_present: bool = None, is_product_code_present: bool = None,
                             run_command: str = None):
    install_manager.is_file_present.return_value = is_file_present
    install_manager.is_product_code_present.return_value = is_product_code_present
    install_manager.run_command.return_value = run_command


def set_download_manager_data(present_paths: Set[str]):
    download_manager.filter_present.side_effect = lambda l: [i for i in l if i.download_path not in present_paths]


class ExtraAssertMixin:

    def assert_called_exactly_with_no_order(self: TestCase, mock, *calls):
        self.assertEqual(mock.call_count, len(calls))
        mock.assert_has_calls(calls, any_order=True)


def assert_download_progress_for(mock: Feedback, expected: Set[DownloadSpec]):
    encountered = set()
    for ci in mock.report_entry_download_progress.call_args_list:
        ((sp, _), _) = ci
        encountered.add(sp)
    if expected != encountered:
        raise AssertionError('No progress for {}'.format(expected))


def _get_test_middleware(log: List):
    @middleware
    async def test_middleware(request, handler):
        log.append((request.raw_path, request.headers.get('Range')))
        return await handler(request)

    return test_middleware


class _TestWebServerCore:

    def __init__(self, host: str, port: int, calls: List, runner: AppRunner, site: TCPSite):
        self.host = host
        self.port = port
        self.calls = calls
        self.runner = runner
        self.site = site

    async def stop(self) -> None:
        await self.runner.cleanup()


_dir_path = Path(__file__).parent / 'test_files'
_chunk = 16


def create_archive(archive_path: Path, archive_format: str):
    return make_archive(archive_path, archive_format, _dir_path)


async def _handle_test4(request):
    resp = StreamResponse()
    with open(_dir_path / 'test4.txt', 'rb') as fo:
        fo.seek(0, SEEK_END)
        resp.content_length = fo.tell()
        fo.seek(0)
        await resp.prepare(request)
        for data in iter(partial(fo.read, _chunk), b''):
            await resp.write(data)
    await resp.write_eof()
    return resp


async def _handle_test5(_):
    return Response(status=HTTPRequestRangeNotSatisfiable.status_code)


async def _handle_test6(_):
    resp = Response(status=HTTPRequestRangeNotSatisfiable.status_code)
    resp.headers['Content-Range'] = 'meters 1-2/3'
    return resp


async def _create_test_web_server_core() -> _TestWebServerCore:
    calls = []
    app = Application(middlewares=[_get_test_middleware(calls)])
    app.router.add_static('/static/', _dir_path)
    app.router.add_get('/dynamic/test4.txt', _handle_test4)
    app.router.add_get('/dynamic/test5.txt', _handle_test5)
    app.router.add_get('/dynamic/test6.txt', _handle_test6)
    runner = AppRunner(app)
    await runner.setup()
    host = 'localhost'
    port = unused_port()
    site = TCPSite(runner, host, port)
    await site.start()
    return _TestWebServerCore(host, port, calls, runner, site)


class TestWebServer:

    def __init__(self):
        self.loop = None
        self.core = None

        def helper():
            loop = new_event_loop()
            set_event_loop(loop)
            self.loop = loop
            self.core = loop.run_until_complete(_create_test_web_server_core())
            self.started.set()
            self.loop.run_forever()
            tasks = all_tasks(self.loop)
            loop.run_until_complete(gather(*tasks, return_exceptions=True))
            loop.run_until_complete(loop.shutdown_asyncgens())
            self.loop.close()
            set_event_loop(None)

        self.started = Event()
        self.thread = Thread(target=helper)

    def start(self):
        self.thread.start()
        self.started.wait()

    def stop(self):
        async def helper():
            await self.core.stop()
            self.loop.stop()

        run_coroutine_threadsafe(helper(), self.loop)
        self.thread.join()

    def create_url_for(self, file_name: str) -> str:
        return urlunsplit(('http', f'{self.core.host}:{self.core.port}', file_name, '', ''))

    def reset_request_data(self):
        self.core.calls.clear()

    def get_request_data(self):
        return list(self.core.calls)


def copy_to(test_file: str, dst: str, n_bytes: int = None):
    with open(Path(__file__).parent / 'test_files' / test_file, 'rb') as sfo:
        with open(dst, 'wb') as dfo:
            if n_bytes:
                full_reads = n_bytes // _chunk
                partial_read = n_bytes % _chunk
                for _ in range(full_reads):
                    data = sfo.read(_chunk)
                    if data:
                        dfo.write(data)
                    else:
                        break
                else:
                    data = sfo.read(partial_read)
                    dfo.write(data)
            else:
                for data in iter(partial(sfo.read, _chunk), b''):
                    dfo.write(data)



# Generated from EnvParser.g4 by ANTLR 4.7.2
# encoding: utf-8
import sys
from io import StringIO

from antlr4 import *
from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\7")
        buf.write("\36\4\2\t\2\4\3\t\3\4\4\t\4\3\2\3\2\3\3\3\3\3\3\3\3\7")
        buf.write("\3\17\n\3\f\3\16\3\22\13\3\3\3\5\3\25\n\3\3\3\3\3\5\3")
        buf.write("\31\n\3\3\4\3\4\3\4\3\4\2\2\5\2\4\6\2\3\3\2\4\5\2\35\2")
        buf.write("\b\3\2\2\2\4\30\3\2\2\2\6\32\3\2\2\2\b\t\t\2\2\2\t\3\3")
        buf.write("\2\2\2\n\31\7\2\2\3\13\20\5\2\2\2\f\r\7\6\2\2\r\17\5\2")
        buf.write("\2\2\16\f\3\2\2\2\17\22\3\2\2\2\20\16\3\2\2\2\20\21\3")
        buf.write("\2\2\2\21\24\3\2\2\2\22\20\3\2\2\2\23\25\7\6\2\2\24\23")
        buf.write("\3\2\2\2\24\25\3\2\2\2\25\26\3\2\2\2\26\27\7\2\2\3\27")
        buf.write("\31\3\2\2\2\30\n\3\2\2\2\30\13\3\2\2\2\31\5\3\2\2\2\32")
        buf.write("\33\5\2\2\2\33\34\7\2\2\3\34\7\3\2\2\2\5\20\24\30")
        return buf.getvalue()


class EnvParser(Parser):
    grammarFileName = "EnvParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = ["<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                    "';'"]

    symbolicNames = ["<INVALID>", "WS", "UNQUOTED_STRING", "DOUBLE_QUOTED_STRING",
                     "DELIM", "UNRECOGNIZED"]

    RULE_entry = 0
    RULE_path = 1
    RULE_single_entry = 2

    ruleNames = ["entry", "path", "single_entry"]

    EOF = Token.EOF
    WS = 1
    UNQUOTED_STRING = 2
    DOUBLE_QUOTED_STRING = 3
    DELIM = 4
    UNRECOGNIZED = 5

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None

    class EntryContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def UNQUOTED_STRING(self):
            return self.getToken(EnvParser.UNQUOTED_STRING, 0)

        def DOUBLE_QUOTED_STRING(self):
            return self.getToken(EnvParser.DOUBLE_QUOTED_STRING, 0)

        def getRuleIndex(self):
            return EnvParser.RULE_entry

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterEntry"):
                listener.enterEntry(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitEntry"):
                listener.exitEntry(self)

    def entry(self):

        localctx = EnvParser.EntryContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_entry)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 6
            _la = self._input.LA(1)
            if not (_la == EnvParser.UNQUOTED_STRING or _la == EnvParser.DOUBLE_QUOTED_STRING):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PathContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(EnvParser.EOF, 0)

        def entry(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(EnvParser.EntryContext)
            else:
                return self.getTypedRuleContext(EnvParser.EntryContext, i)

        def DELIM(self, i: int = None):
            if i is None:
                return self.getTokens(EnvParser.DELIM)
            else:
                return self.getToken(EnvParser.DELIM, i)

        def getRuleIndex(self):
            return EnvParser.RULE_path

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPath"):
                listener.enterPath(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPath"):
                listener.exitPath(self)

    def path(self):

        localctx = EnvParser.PathContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_path)
        self._la = 0  # Token type
        try:
            self.state = 22
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [EnvParser.EOF]:
                self.enterOuterAlt(localctx, 1)
                self.state = 8
                self.match(EnvParser.EOF)
                pass
            elif token in [EnvParser.UNQUOTED_STRING, EnvParser.DOUBLE_QUOTED_STRING]:
                self.enterOuterAlt(localctx, 2)
                self.state = 9
                self.entry()
                self.state = 14
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 0, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 10
                        self.match(EnvParser.DELIM)
                        self.state = 11
                        self.entry()
                    self.state = 16
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 0, self._ctx)

                self.state = 18
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == EnvParser.DELIM:
                    self.state = 17
                    self.match(EnvParser.DELIM)

                self.state = 20
                self.match(EnvParser.EOF)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Single_entryContext(ParserRuleContext):

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def entry(self):
            return self.getTypedRuleContext(EnvParser.EntryContext, 0)

        def EOF(self):
            return self.getToken(EnvParser.EOF, 0)

        def getRuleIndex(self):
            return EnvParser.RULE_single_entry

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSingle_entry"):
                listener.enterSingle_entry(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSingle_entry"):
                listener.exitSingle_entry(self)

    def single_entry(self):

        localctx = EnvParser.Single_entryContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_single_entry)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 24
            self.entry()
            self.state = 25
            self.match(EnvParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import namedtuple
from unittest import TestCase
from unittest.mock import NonCallableMock

from thespiae.conf import AppData
from thespiae.conf.core import create_app_entry
from thespiae.path.core import get_path_entries, get_single_path_entry, construct_new_user_path
from thespiae.path.exception import PathEntryError, PathError, InconsistentUserPathExtensionError, \
    InconsistentPathEntryExtensionError, InconsistentUserPathEntryDataError, InconsistentActualPathEntryDataError, \
    InvalidPathEntryError, InvalidSystemPathError, InvalidUserPathError
from thespiae.path.protocol import Feedback
from .helper import complex_string, PathManagerMockResetMixin, set_path_manager_data
from .singleton import path_processor, path_manager


class PathConstructorTest(TestCase):

    def test_getting_path_entries(self):
        empty = ('', ([], []))
        strings = ('   "" ;"  " ', (['""', '"  "'], ['', '  ']))
        cx = (complex_string, (['"a bc"', 'def', "'g", "h'"], ['a bc', 'def', "'g", "h'"]))
        for path_string, (expected_raw_entries, expected_entries) in [empty, strings, cx]:
            with self.subTest(path_string=path_string, expected_raw_entries=expected_raw_entries,
                              expected_entries=expected_entries):
                self.assertSequenceEqual(get_path_entries(path_string, True), expected_raw_entries)
                self.assertSequenceEqual(get_path_entries(path_string, False), expected_entries)

    def test_getting_path_entries2(self):
        with self.assertRaises(PathError):
            get_path_entries('";', True)

    def test_getting_path_entries3(self):
        with self.assertRaises(PathError):
            get_path_entries('abc;d"e', True)

    def test_gettings_single_path_entry(self):
        self.assertEqual(get_single_path_entry('"abc"', True), '"abc"')
        self.assertEqual(get_single_path_entry('"abc"', False), 'abc')

    def test_gettings_single_path_entry2(self):
        with self.assertRaises(PathEntryError):
            get_single_path_entry('abc;d', True)

    def test_constructing_new_user_path(self):
        dataset = namedtuple('dataset', ['system_path_entries', 'raw_user_path_entries', 'user_path_entries',
                                         'raw_actual_entries', 'actual_entries', 'obsolete_entries'])
        input_data = [dataset(system_path_entries=['C:\\bin'], raw_user_path_entries=[], user_path_entries=[],
                              raw_actual_entries=['"D:\\bi;n"'], actual_entries=['D:\\bi;n'], obsolete_entries=[]),
                      dataset(system_path_entries=['C:\\bin'], raw_user_path_entries=['D:\\bin'],
                              user_path_entries=['D:\\bin'], raw_actual_entries=['D:\\bin'], actual_entries=['D:\\bin'],
                              obsolete_entries=[]),
                      dataset(system_path_entries=['C:\\bin'], raw_user_path_entries=['"D:\\bin"'],
                              user_path_entries=['D:\\bin'], raw_actual_entries=['E:\\bin'], actual_entries=['E:\\bin'],
                              obsolete_entries=[]),
                      dataset(system_path_entries=['C:\\bin'], raw_user_path_entries=['"D:\\bin"'],
                              user_path_entries=['D:\\bin'], raw_actual_entries=['E:\\bin'], actual_entries=['E:\\bin'],
                              obsolete_entries=['D:\\bin']),
                      dataset(system_path_entries=['C:\\bin'], raw_user_path_entries=['"D:\\bin"'],
                              user_path_entries=['D:\\bin'], raw_actual_entries=['E:\\bin'], actual_entries=['E:\\bin'],
                              obsolete_entries=['C:\\bin'])]
        expected_output = ['"D:\\bi;n"', 'D:\\bin', '"D:\\bin";E:\\bin', 'E:\\bin', '"D:\\bin";E:\\bin']
        for inputs, output in zip(input_data, expected_output):
            with self.subTest(**inputs._asdict(), expected=output):
                self.assertEqual(construct_new_user_path(**inputs._asdict()), output)

    def test_constructing_new_user_path2(self):
        with self.assertRaises(InconsistentUserPathEntryDataError):
            construct_new_user_path([], ['C:\\bin'], [], [], [], [])

    def test_constructing_new_user_path3(self):
        with self.assertRaises(InconsistentActualPathEntryDataError):
            construct_new_user_path([], [], [], ['C:\\bin'], [], [])


class PathProcessorTest(PathManagerMockResetMixin, TestCase):

    def test_processing_path_changes(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_path_manager_data('C:\\bin', '"D:\\%TEST_VAR2%n";E:\\bin', {'TEST_VAR': 'i;n', 'TEST_VAR2': 'b;i'})
        app_data = AppData([create_app_entry([{'name': 'a', 'version': '1',
                                               'path_entries': ['"E:\\bin"', '"F:\\b%TEST_VAR%"'], 'keep': True}], []),
                            create_app_entry([{'name': 'b', 'version': '2',
                                               'path_entries': ['"D:\\%TEST_VAR2%n"']}], [])])
        path_processor.process_path_changes(app_data, fb)
        fb.report_path_analysis.assert_called_once()
        fb.confirm_user_path_update.assert_called_once_with('"D:\\%TEST_VAR2%n";E:\\bin', 'E:\\bin;"F:\\b%TEST_VAR%"')
        path_manager.set_user_path.assert_called_once_with('E:\\bin;"F:\\b%TEST_VAR%"')
        path_manager.notify_path_change.assert_called_once()
        fb.report_user_path_updated.assert_called_once()
        fb.report_user_path_no_changes.assert_not_called()

    def test_processing_path_changes2(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_path_manager_data('"C:\\b;in', 'D:\\bi%TEST_VAR%:\\bin', {'TEST_VAR': 'n;E'})
        app_data = AppData([create_app_entry([{'name': 'a', 'version': '1',
                                               'path_entries': ['F:\\bin'], 'keep': True}], []),
                            create_app_entry([{'name': 'b', 'version': '2'}], [])])
        with self.assertRaises(InvalidSystemPathError):
            path_processor.process_path_changes(app_data, fb)
        fb.report_path_analysis.assert_called_once()
        fb.confirm_user_path_update.assert_not_called()
        path_manager.set_user_path.assert_not_called()
        path_manager.notify_path_change.assert_not_called()
        fb.report_user_path_updated.assert_not_called()
        fb.report_user_path_no_changes.assert_not_called()

    def test_processing_path_changes3(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_path_manager_data('C:\\bin', 'D:\\bi%TEST_VAR%:\\b"in', {'TEST_VAR': 'n;E'})
        app_data = AppData([create_app_entry([{'name': 'a', 'version': '1',
                                               'path_entries': ['F:\\bin'], 'keep': True}], []),
                            create_app_entry([{'name': 'b', 'version': '2'}], [])])
        with self.assertRaises(InvalidUserPathError):
            path_processor.process_path_changes(app_data, fb)
        fb.report_path_analysis.assert_called_once()
        fb.confirm_user_path_update.assert_not_called()
        fb.report_path_analysis.assert_called_once()
        path_manager.set_user_path.assert_not_called()
        path_manager.notify_path_change.assert_not_called()
        fb.report_user_path_updated.assert_not_called()
        fb.report_user_path_no_changes.assert_not_called()

    def test_processing_path_changes4(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_path_manager_data('C:\\bin', 'D:\\bi%TEST_VAR%:\\bin', {'TEST_VAR': 'n;E'})
        app_data = AppData([create_app_entry([{'name': 'a', 'version': '1',
                                               'path_entries': ['F:\\bin'], 'keep': True}], []),
                            create_app_entry([{'name': 'b', 'version': '2'}], [])])
        with self.assertRaises(InconsistentUserPathExtensionError) as c:
            path_processor.process_path_changes(app_data, fb)
        fb.report_path_analysis.assert_called_once()
        fb.confirm_user_path_update.assert_not_called()
        path_manager.set_user_path.assert_not_called()
        path_manager.notify_path_change.assert_not_called()
        fb.report_user_path_updated.assert_not_called()
        fb.report_user_path_no_changes.assert_not_called()
        self.assertEqual(c.exception.data, 'D:\\bi%TEST_VAR%:\\bin')
        self.assertEqual(c.exception.extended_data, 'D:\\bin;E:\\bin')

    def test_processing_path_changes5(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_path_manager_data('C:\\bin', 'D:\\bin;E:\\bin', {'TEST_VAR': 'i;n'})
        app_data = AppData([create_app_entry([{'name': 'a', 'version': '1',
                                               'path_entries': ['F:\\b%TEST_VAR%'], 'keep': True}], []),
                            create_app_entry([{'name': 'b', 'version': '2'}], [])])
        with self.assertRaises(InconsistentPathEntryExtensionError) as c:
            path_processor.process_path_changes(app_data, fb)
        fb.report_path_analysis.assert_called_once()
        fb.confirm_user_path_update.assert_not_called()
        path_manager.set_user_path.assert_not_called()
        path_manager.notify_path_change.assert_not_called()
        fb.report_user_path_updated.assert_not_called()
        fb.report_user_path_no_changes.assert_not_called()
        self.assertEqual(c.exception.data, 'F:\\b%TEST_VAR%')
        self.assertEqual(c.exception.extended_data, 'F:\\bi;n')

    def test_processing_path_changes6(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_path_manager_data('C:\\bin', 'D:\\bin;E:\\bin', {})
        app_data = AppData([create_app_entry([{'name': 'a', 'version': '1',
                                               'path_entries': ['F:\\bi;n'], 'keep': True}], []),
                            create_app_entry([{'name': 'b', 'version': '2'}], [])])
        with self.assertRaises(InvalidPathEntryError) as c:
            path_processor.process_path_changes(app_data, fb)
        fb.report_path_analysis.assert_called_once()
        fb.confirm_user_path_update.assert_not_called()
        self.assertEqual(c.exception.entry, 'F:\\bi;n')
        self.assertEqual(c.exception.app_entry, create_app_entry([{'name': 'a', 'version': '1',
                                                                   'path_entries': ['F:\\bi;n'], 'keep': True}], []))
        fb.report_user_path_updated.assert_not_called()
        fb.report_user_path_no_changes.assert_not_called()
        path_manager.set_user_path.assert_not_called()
        path_manager.notify_path_change.assert_not_called()

    def test_processing_path_changes7(self):
        fb = NonCallableMock(spec_set=Feedback)
        set_path_manager_data('C:\\bin', '"D:\\%TEST_VAR2%n";E:\\bin', {'TEST_VAR': 'i;n', 'TEST_VAR2': 'b;i'})
        app_data = AppData([create_app_entry([{'name': 'a', 'version': '1', 'path_entries': [], 'keep': True}], []),
                            create_app_entry([{'name': 'b', 'version': '2', 'path_entries': []}], [])])
        path_processor.process_path_changes(app_data, fb)
        fb.report_path_analysis.assert_called_once()
        fb.confirm_user_path_update.assert_not_called()
        path_manager.set_user_path.assert_not_called()
        path_manager.notify_path_change.assert_not_called()
        fb.report_user_path_updated.assert_not_called()
        fb.report_user_path_no_changes.assert_called_once()
